/***************************************************************
 * Name:      audio.h
 * Purpose:   audio streams raw waveforms to audio
 * Author:    John Conner Atkeson
 * Created:   2007-11-06
 * Copyright: John Atkeson (http://jca3.freeshell.org)
 * License:   LPGL
 *  code heavily derived from:
 *  https://lampka.siedziba.pl/WebSVN/filedetails.php?repname=siedziba+svn&path=%2FSound_Test%2FSoundTest.cpp&rev=256&sc=1
 **************************************************************/
#include "audio.h"
// https://lampka.siedziba.pl/WebSVN/filedetails.php?repname=siedziba+svn&path=%2FSound_Test%2FSoundTest.cpp&rev=256&sc=1
#include "ctype.h"
#include "SDL/SDL.h"//-lSDLmain -lSDL -mwindow
#include "SDL/SDL_audio.h"
#include <math.h>
#include <stdio.h>
#ifdef WINDOWS
#include "windows.h"
#endif

#ifdef WINDOWS
#define die(...) { char err[100]; snprintf(err, sizeof(err), __VA_ARGS__); MessageBox(NULL, err, "Error", MB_OK | MB_ICONEXCLAMATION); exit(1); }
#else
#define die(...) { fprintf(stderr, __VA_ARGS__); fprintf(stderr, "\n"); exit(1); }
#endif
#include <windows.h>

#include <pthread.h>

int num_iterations = 1;

/*  Make the class name into a global variable  */
// char szClassName[ ] = "CodeBlocksWindowsApp";

//unsigned             Initialised = 0;

// User controllable variables
// const unsigned  SAMPLE_FREQ = 44100;     // sampling frequency
// unsigned        sdl_samples = 2048;          // SDL buffer size

/* ************************************************************************************************************ */
/* Playable_Class is the base class for every object that can output a waveform. */
Playable_Class::Playable_Class(){
  parent=NULL;
  time=0.0; this->set_octave(7.0); loudness=1.0;
};
void Playable_Class::set_octave(double new_octave){
  octave = new_octave;
  frequency = Math.Pow(2.0, new_octave);
};
void Playable_Class::set_frequency(double new_frequency){
  octave = Math.Log(new_frequency)/Math.Log(2.0);
  frequency = new_frequency;
};
double Playable_Class::getamp(double time){
  return 0.0;
};
double Playable_Class::maxamp(){
  return 1.0;
};
double Playable_Class::get_duration(){// tells how long this track will play
  return 1.0;
};
bool Playable_Class::keepgoing(){// tells when we have passed the end of duration.
  return true;
};
//-------------------------------------------------------------------------------------
void Playable_Class::Get_Base_Abs(double *xorg, double *yorg, double *zorg){
  /* return the absolute coordinates of my origin */
  double xloc=0.0, yloc=0.0, zloc=1.0;
  Playable_Class *pptr = this->parent;
  while (pptr!=NULL){// reach up through all tree parents
    xloc += pptr->time;
    yloc += pptr->octave;
    zloc *= pptr->loudness;// multiply loudness, envelopes within envelopes
    pptr = pptr->parent;
  }
  (*xorg)=xloc; (*yorg)=yloc; (*zorg)=zloc;
};/* Playable_Class::Get_Base_Abs */
class Time_Slicer_Class* Playable_Class::spawn_slicer(){// creates a "read head" to keep a time marker in this playable
  return new Time_Slicer_Class(this);
};
/* **********************************************************************************************************************
  Begin Playable_Tree_Class functions.
********************************************************************************************************************** */
ArrayList<Playable_Class> *kids;
Playable_Tree_Class::Playable_Tree_Class(){
  kids = new ArrayList<Playable_Class>();
  time=0.0; octave=0.0; loudness=1.0;
};
Playable_Tree_Class::~Playable_Tree_Class(){
  delete kids;
};
void Playable_Tree_Class::Auto_Add_Kids(int num_kids){
  for (int cnt=0;cnt<num_kids;cnt++){
    Signal_Generator_Class *sgen = new Signal_Generator_Class();
    kids->add(new Voice_Track_Class(sgen));
  }
};
void Playable_Tree_Class::Auto_Remove_Kids(){
  int num_kids = kids->size();
  for (int cnt=0;cnt<num_kids;cnt++){
    Voice_Track_Class *vt = (Voice_Track_Class*)(kids->get(cnt));
    Signal_Generator_Class *sgen = vt->sgen;
    delete vt;
    delete sgen;
  }
};
overrider double Playable_Tree_Class::getamp(double time){
  return 0.0;
};
overrider double Playable_Tree_Class::maxamp(){
  return 1.0;
};
overrider double Playable_Tree_Class::get_duration(){
  return 0.0;
};// tells how long this track will play
overrider bool Playable_Tree_Class::keepgoing(){
  return true;
};// tells when we have passed the end of duration.
overrider class Time_Slicer_Class* Playable_Tree_Class::spawn_slicer(){// creates a "read head" to keep a time marker in this playable
  return new Time_Slicer_Tree_Class(this);
};
// Drawable_Class functions
overrider void Playable_Tree_Class::Draw_Me(wxDC& dc, Box_Mapper_Class *Box_Map){
  int num_kids = this->kids->size();
  for (int cnt=0;cnt<num_kids;cnt++){
    Playable_Class *kid = kids->get(cnt);
    kid->Draw_Me(dc, Box_Map);
  }
};
overrider class Playable_Class* Playable_Tree_Class::Deliver_Playable(double xloc,double yloc, Box_Mapper_Class *Box_Map){
  // borrow this from Voice_Track_Class;
  return NULL;
};
overrider bool Playable_Tree_Class::touchedme(int mousex,int mousey){ return false; };
overrider bool Playable_Tree_Class::touched_my_handle(int mousex,int mousey){ return false; };
/* **********************************************************************************************************************
  Begin Time_Slicer_Class functions.
********************************************************************************************************************** */
Time_Slicer_Class::Time_Slicer_Class(class Playable_Class *new_playable){
  my_playable = new_playable;
  time_base_abs=0.0; octave_base_abs=0.0; loudness_base_abs=0.0;
  time_abs=0.0; octave_abs=0.0; loudness_abs=0.0;
  parent = NULL;
  keeprunning = true;
  this->Establish_Location();
};
//-------------------------------------------------------------------------------------
Time_Slicer_Class::~Time_Slicer_Class(){};
//-------------------------------------------------------------------------------------
void Time_Slicer_Class::init_slicer(double start_time){// this one is different for every descendent
  Establish_Location();
};
//-------------------------------------------------------------------------------------
double Time_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  double returnval = 0.0;
  nowtime -= my_playable->time; nowoctave += my_playable->octave;/* yloc is base octave for this frame */ nowvol *= my_playable->loudness;
  returnval = my_playable->getamp(nowtime);
  return returnval;
};
//-------------------------------------------------------------------------------------
bool Time_Slicer_Class::keepgoing(){// tells when we have passed the end of duration.
  return keeprunning;
};
//-------------------------------------------------------------------------------------
void Time_Slicer_Class::Establish_Location(){// reestablish my absolute coordinates based on parents
  /* This is called: 1. when it is spawned for playing, and 2. whenever my playable or any of its parents are moved. */
  Time_Slicer_Class *pptr = this->parent;
  double xloc, yloc, zloc;
  if (true){
    this->my_playable->Get_Base_Abs(&xloc, &yloc, &zloc);
  }else{
    //xloc=my_playable->time; yloc=my_playable->octave; zloc=my_playable->loudness;
    xloc=0.0; yloc=0.0; zloc=1.0;
    Playable_Class *playable;
    while (pptr!=NULL){// reach up through all tree parents
      playable = pptr->my_playable;
      xloc += playable->time;
      yloc += playable->octave;
      zloc *= playable->loudness;// multiply loudness, envelopes within envelopes
      pptr = pptr->parent;
    }
  }
  this->time_base_abs=xloc; this->octave_base_abs=yloc; this->loudness_base_abs=zloc;
  this->time_abs=xloc+my_playable->time;
  this->octave_abs=yloc+my_playable->octave;
  this->loudness_abs=zloc*my_playable->loudness;
  // this->time_abs=xloc; this->octave_abs=yloc; this->loudness_abs=zloc;
};
/* **********************************************************************************************************************
  Begin Time_Slicer_Tree_Class functions.
********************************************************************************************************************** */
Time_Slicer_Tree_Class::Time_Slicer_Tree_Class(class Playable_Tree_Class *new_playable):Time_Slicer_Class(new_playable){
  subslicers = new ArrayList<Time_Slicer_Class>();
  my_playable_tree = new_playable;
};
//-------------------------------------------------------------------------------------
//overrider void Time_Slicer_Tree_Class::Establish_Location();// reestablish my absolute coordinates based on parents
//-------------------------------------------------------------------------------------
overrider Time_Slicer_Tree_Class::~Time_Slicer_Tree_Class(){
  delete subslicers;
};
//-------------------------------------------------------------------------------------
overrider void Time_Slicer_Tree_Class::init_slicer(double start_time){// this one is different for every descendent
  int num_subslices = subslicers->size();
  for (int scnt=0;scnt<num_subslices;scnt++){
    Time_Slicer_Class *slice = subslicers->get(scnt);
    slice->init_slicer(start_time);
  }
};
//-------------------------------------------------------------------------------------
overrider double Time_Slicer_Tree_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  double returnval = 0.0;
  nowtime -= my_playable->time; nowoctave += my_playable->octave;/* yloc is base octave for this frame */ nowvol *= my_playable->loudness;
  int num_subslices = subslicers->size();
  for (int scnt=0;scnt<num_subslices;scnt++){
    Time_Slicer_Class *slice = subslicers->get(scnt);
    returnval += slice->getslice(nowtime, nowoctave, nowvol);
  }
  returnval /= (double)num_subslices;
  // if we have passed the end of my playable's duration, set keepgoing to false.
  // keepgoing = my_playable->keepgoing();
  return returnval;
};
//-------------------------------------------------------------------------------------
overrider bool Time_Slicer_Tree_Class::keepgoing(){// tells when we have passed the end of duration.
  return keeprunning;
};
/* **********************************************************************************************************************
  Begin Player_Class functions and callback
********************************************************************************************************************** */
void player_callback_wrapper(void *environment, Uint8 *stream, int len){
  Player_Class *player = (Player_Class*)environment;  stereo *fill = (stereo*)stream;
  // player->fill_outbuffer(fill, len/sizeof(stereo));
  player->fill_outbuffer_sgen(fill, len/sizeof(stereo));
};
//-------------------------------------------------------------------------------------
void Player_Class::sdlsoundinit(){
  // initialise the SDL system
  // sdlinit(); return;
  if (false){
    if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_VIDEO) == -1){
      die("Could not initialize SDL: %s", SDL_GetError());
    }
  }else{
    /* Initialization */
    if (SDL_InitSubSystem(SDL_INIT_AUDIO) == -1){
      // SDL Audio subsystem could not be started
      die("Could not initialize SDL: %s", SDL_GetError());
    }
  }
  atexit(SDL_Quit);
  if (Initialised) SDL_CloseAudio();
  SDL_AudioSpec format;
  format.format = AUDIO_S16SYS;
  format.channels = 2;
  format.freq = SAMPLE_FREQ;
  format.samples = sdl_samples;
  format.callback = player_callback_wrapper;
  format.userdata = this;
  if (SDL_OpenAudio(&format, NULL) < 0) die("SDL_OpenAudio failed: %s", SDL_GetError());
  SDL_PauseAudio(0);
  Initialised = 1;
};
//-------------------------------------------------------------------------------------
int Player_Class::fill_outbuffer(stereo *fill, int playsize){// virtual
#if 0
  float sine_step = Frequency * twopi / SAMPLE_FREQ;
  while (playsize > 0) // write into the sound buffer
  {
    CyclePos += sine_step;
    if (CyclePos > twopi) CyclePos -= twopi;
    fill->left = fill->right = (short)nearbyintf(12000 * sinf(CyclePos));
    fill++;
    playsize --;
  }
#endif
  return 0;
};
//-------------------------------------------------------------------------------------
int Player_Class::fill_outbuffer_sgen(stereo *fill, int playsize){
  while (playsize > 0) // write into the sound buffer
  {
    //for (int cnt=0;cnt<num_iterations;cnt++)// slowtest
    {
      //fill->left = fill->right = (short)nearbyintf(12000 * slicer->getslice(time, 0.0, 1.0));
      fill->left = fill->right = (short)nearbyintf(32768 * slicer->getslice(time, 0.0, 1.0));// 2^15 = 32,768
    }
    time+=timeinc;
    fill++;
    playsize--;
  }
  return 0;
};
#if 0
/*
next step is to figure out how to play a Voice_Track_Class (voice) in a loop.;
that is a subset of the Voice_Track_Class, startdex to enddex.;

at some point we want more than sine waves;
create a gui freewave_generator_class editor.;

ideally the player defines the time range to play in a loop,;
and ideally the current time must only be expressed to the playables as an absolute time.;

how does the player know the end-of-loop time when that time is moving?;
*/
#endif
/* **********************************************************************************************************************
  End Player_Class methods and callback
********************************************************************************************************************** */
/* **********************************************************************************************************************
  Begin Signal_Generator_Class methods
********************************************************************************************************************** */
Signal_Generator_Class::Signal_Generator_Class(){
  loudness = 1.0;
  // freq = 261.626;// default to approximately middle C
  freq = 440.0;// default to A, = octave 8.7813597135246596040696824762296
  clear();
}
//-------------------------------------------------------------------------------------
double Signal_Generator_Class::getamp(double time){// time is in absolute seconds from T=0.
  calc_angle(time);
  return Math.Sin(current_angle * twopi) * loudness;// twopi == 360 degrees (radians)
}
//-------------------------------------------------------------------------------------
double Signal_Generator_Class::maxamp(){
  return 1.0;// should be 1.0
}
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::clear(){
  reset_time(0.0); current_angle = 0.0;
}
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::reset_time(double new_time){
  prev_time = new_time;//
}
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::set_octave(double new_octave){
  octave = new_octave;
  freq = Math.Pow(2.0, new_octave);
};
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::set_frequency(double new_freq){
  freq = new_freq;
  octave = Math.Log(new_freq)/Math.Log(2.0);
  //set_octave(Math.Log(new_freq)/Math.Log(2.0)); // alternative
};
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::calc_angle(double time){
  double delta_time = time-prev_time;
  double delta_angle = delta_time * freq;// 1.0 == 360 degrees (fractions along the way)
  current_angle += delta_angle;// current_angle = current_angle-Math.floor(current_angle);
  prev_time = time;
};
/* **********************************************************************************************************************
  End Signal_Generator_Class methods
********************************************************************************************************************** */

/* **********************************************************************************************************************
  Begin Freewave_Generator_Class methods
********************************************************************************************************************** */
Freewave_Generator_Class::Freewave_Generator_Class(double new_freq){
  rez = 1000;
  //rez = 10000;
  frez = (double)rez;// resolution of our waveform
  ampray = allocsafe(rez, double);
  freq = new_freq;
#if 0
  for (int cnt=0;cnt<rez;cnt++){
    ampray[cnt] = (  2.0*(((double)cnt)/frez)  )-1.0;// sawtooth wave
  }
#else
  for (int cnt=0;cnt<rez;cnt++){
    ampray[cnt] = Math.Sin(twopi*((double)cnt)/frez);// sine wave
  }
#endif
};
//-------------------------------------------------------------------------------------
Freewave_Generator_Class::~Freewave_Generator_Class(){
  freesafe(ampray);
};
//-------------------------------------------------------------------------------------
overrider double Freewave_Generator_Class::getamp(double time){// time is in absolute seconds from T=0.
  calc_angle(time);
  double base_angle = current_angle;
  double base = Math.Floor(base_angle); base_angle -= base;
  base_angle *= frez;

  // get the two array indexes that surround us
  int dex0 = (int)base_angle; // int dex2 = (dex+1) % 100;
  int dex1 = (dex0+1); if (dex1>=rez) {
    dex1=0;
  }// wrap to beginning

  // interpolate between points.
  double weight = base_angle; base = Math.Floor(weight); weight -= base;
  double value = (ampray[dex0]*(1.0-weight)) + (ampray[dex1]*weight);
  return value * loudness;
};
#if 0
//-------------------------------------------------------------------------------------
double Freewave_Generator_Class::find_by_time(double time){// time is in absolute seconds from T=0.
  calc_angle(time);
  double base_angle = current_angle;
  double base = Math.Floor(base_angle); base_angle -= base;
  base_angle *= frez;

  // get the two array indexes that surround us
  int dex0 = (int)base_angle; // int dex2 = (dex+1) % 100;
  int dex1 = (dex0+1);
  if (dex1>=rez) {
    dex1=0;
  }// wrap to beginning

  // interpolate between points.
  double weight = base_angle; base = Math.Floor(weight); weight -= base;
  double value = (ampray[dex0]*weight) + (ampray[dex1]*(1.0-weight));
  return value * loudness;
};
#endif

/* **********************************************************************************************************************
  Begin Splinewave_Generator_Class methods
********************************************************************************************************************** */
Splinewave_Generator_Class::Splinewave_Generator_Class(double new_freq):Freewave_Generator_Class(new_freq)
{
  Point_Class *pnt;
  spline = new ArrayList<Point_Class>();
  // xp0 = 2.2; xp1 = 12.3; xp2 = 15.6; yp0 = 2.2; yp1 = 12.3; yp2 = 15.6;
  switch (5){
  case 0:{
    pnt = new Point_Class(0.22,  0.22); spline->add(pnt);
    pnt = new Point_Class(0.46, -0.90); spline->add(pnt);
    pnt = new Point_Class(0.53,  0.93); spline->add(pnt);
    pnt = new Point_Class(0.76, -0.40); spline->add(pnt);
    break;
  }
  case 1:{// pointy wave
    pnt = new Point_Class(0.22, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.45,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.55,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 2:{// has gaps
    pnt = new Point_Class(0.20, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.40,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.60,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 3:{// symmetric diving bell
    pnt = new Point_Class(0.20, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.30,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.70,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 4:{
    pnt = new Point_Class(0.01, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.30,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.50,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.70,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.75, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 5:{//senseless loop
    double num=16;
    for (double cnt=0;cnt<num;cnt++){
      double yval = Math.Sin(twopi * cnt/num);
      double jitter = Math.Frand()*0.00001;
      pnt = new Point_Class(jitter + cnt/num, yval); spline->add(pnt);
    }
    break;
  }
  }/* switch */
  apply_spline();
};
//-------------------------------------------------------------------------------------
void Splinewave_Generator_Class::apply_spline(){// apply the spline to the amplitude array
  Point_Class *first_point, *prev_point, *next_point;
  int num_points = spline->size();

  first_point = spline->get(num_points-2);// last_point-1

  double xfirst,yfirst;
  xfirst = first_point->xloc-1.0; yfirst = first_point->yloc;// wrap

  double xprev,yprev, xnext,ynext;
  prev_point = spline->get(num_points-1);// last_point
  xprev = prev_point->xloc-1.0; yprev = prev_point->yloc;// wrap

  double xcut_prev, ycut_prev, xcut_next, ycut_next;
  xcut_prev = (xfirst+xprev)/2; ycut_prev = (yfirst+yprev)/2;
  xcut_next = (xprev+xnext)/2; ycut_next = (yprev+ynext)/2;

  double xtime=0.0;
  double yval=0.0;

  int timedex = (int)(xcut_prev*frez);// increment by integer to avoid cumulative error
  int raydex = (int)((xcut_prev-Math.Floor(xcut_prev))*frez);
  xtime = ((double)timedex)/frez;
  for (int pcnt=0;pcnt<num_points;pcnt++){
    next_point = spline->get(pcnt); xnext = next_point->xloc; ynext = next_point->yloc;
    xcut_next = (xprev+xnext)/2; ycut_next = (yprev+ynext)/2;
    while (xtime<=xcut_next){
      yval = spline_amp(xtime, xcut_prev,ycut_prev, xprev,yprev, xcut_next,ycut_next);
      this->ampray[raydex % rez] = yval;
      raydex++;
      timedex++; xtime = ((double)timedex)/frez;
    }
    xcut_prev = xcut_next; ycut_prev = ycut_next;
    xprev = xnext; yprev = ynext;
  }
};/* apply_spline */
//-------------------------------------------------------------------------------------
double Splinewave_Generator_Class::spline_amp(double X, double xp0, double yp0, double xp1, double yp1, double xp2, double yp2)
{// spline algerbra computed by http://www.quickmath.com/
  double Y, t, comp_t, numer, denom;
  denom = (xp0 + xp2 - (2.0 * xp1));
  if (denom==0.0) {
    xp0+=fudge;// xp0+=0.00000000000001;
    denom = fudge;
  }
  numer = (xp0 - xp1 + Math.Sqrt((xp1 * xp1) - (2.0 * X * xp1) - (xp0 * xp2) + (xp0 * X) + (xp2 * X)));
  t = numer / denom;
  comp_t = 1.0 - t;
  Y = (comp_t * comp_t * yp0) + (2.0 * t * comp_t * yp1) + (t * t * yp2);
  return Y;
};
/* **********************************************************************************************************************
  Begin Voice_Track_Class methods
********************************************************************************************************************** */
Voice_Track_Class::Voice_Track_Class(Signal_Generator_Class *new_sgen){// constructor
  length = 0.0;
  // sgen = new Signal_Generator_Class();
  sgen = new_sgen;//new Freewave_Generator_Class(220.0);
  // I guess we invoke this in the drawing_canvas;
  //fill_test();
  handle_radius = 7;
};
//-------------------------------------------------------------------------------------
void Voice_Track_Class::Set_Signal_Generator(Signal_Generator_Class *new_sgen){
  this->sgen = new_sgen;//Reffy_Ptr_Class sgen_ptr;
}
//-------------------------------------------------------------------------------------
Voice_Track_Class::~Voice_Track_Class(){// destructor
  int num_events = this->size();
  for (int cnt=0; cnt<num_events;cnt++){
    Event_Marker_Class *event = this->get(cnt);
    delete event;
  }
  // voice track is not responsible for deleting its own signal generator.
};
//-------------------------------------------------------------------------------------
overrider void Voice_Track_Class::add(Event_Marker_Class *new_member){
  ArrayList<Event_Marker_Class>::add(new_member);
  new_member->my_voice = this;// connect back to parent
  new_member->parent = this;
  //wxMouseEventHandler(Waveform_Canvas_Class::OnDrawingPanelLeftDown));
  //(Voice_Track_Class::)(Event_Marker_Class*) fred = Voice_Track_Class::add;
  // void (Voice_Track_Class::)(Event_Marker_Class*) fred = Voice_Track_Class::add;
  //void *fred = Voice_Track_Class::add;
};
//-------------------------------------------------------------------------------------
int Voice_Track_Class::find_by_time(double target_time){// given the time, treesearch and return the closest index
  int retval = 0;
  int num_events = this->size();
  int last_event = num_events-1;
  Event_Marker_Class *event_prev = this->get(0);
  Event_Marker_Class *event_next = NULL;
  int cnt = 0;
  while (cnt<last_event){
    cnt++;
    event_next = this->get(cnt);
    if (event_next->time > target_time){
      break;
    }
    event_prev = event_next;
  }
  if (event_next!=NULL){
    double diff0 = fabs(event_prev->time - target_time);
    double diff1 = fabs(event_next->time - target_time);
    if (diff0<diff1){// closer to prev
      retval = cnt-1;
    }else{
      retval = cnt;
    }
  }else{
    retval = cnt-1;
  }
  printf("timint:%i\n",retval);
  return retval; // taskkill /F /IM fretless.exe
#if 0 // treesearch
  double targetval, minval, medval, maxval;
  int mindex=0, meddex, maxdex=size();
  minval = this->get(mindex);
  while (mindex<maxdex){
    meddex = (mindex+maxdex)/2; medval = this->get(meddex);
    if (targetval<=medval){
      maxdex = meddex;
    }else{
      mindex = meddex+1; minval = this->get(mindex);
    }
  }
  return mindex;
#endif
};/* find_by_time */
//-------------------------------------------------------------------------------------
void Voice_Track_Class::fill_test(){// fill Voice_Track with dummy values
  double offset = 0.2;
  double octave_offset = 7.0;
  double freq_offset = 220;
  this->set_octave(octave_offset);
  double fo_220 = 7.7813597135246608 - octave_offset;
  double fo_440 = 8.78135971352466 - octave_offset;
  double fo_880 = 9.78135971352466 - octave_offset;
  /*
  Math.Log(220)/Math.Log(2.0) 7.7813597135246608
  Math.Log(440)/Math.Log(2.0) 8.78135971352466
  Math.Log(880)/Math.Log(2.0) 9.78135971352466
  */
  Event_Marker_Class *event = new Event_Marker_Class();
  event->time = offset+0.20;
  //event->set_octave(Math.Log(freq_offset+220)/Math.Log(2.0));
  event->set_octave(fo_440);
  this->add(event);

  event = new Event_Marker_Class();
  event->time = offset+0.5;
  //event->set_octave(Math.Log(freq_offset+220)/Math.Log(2.0));
  event->set_octave(fo_220);

  event->loudness = 0.5;
  this->add(event);

  event = new Event_Marker_Class();
  event->time = offset+1.0;
  //event->set_octave(Math.Log(freq_offset+220)/Math.Log(2.0));
  event->set_octave(fo_440);
  this->add(event);

  event = new Event_Marker_Class();
  event->time = offset+1.5;
  //event->set_octave(Math.Log(freq_offset+220)/Math.Log(2.0));
  event->set_octave(fo_440);
  this->add(event);

  event = new Event_Marker_Class();
  event->time = offset+3.0;
  //event->set_octave(Math.Log(freq_offset+660)/Math.Log(2.0));
  event->set_octave(fo_880);
  event->loudness = 0.0;
  this->add(event);
  length = event->time;
};
//-------------------------------------------------------------------------------------
overrider double Voice_Track_Class::get_duration(){// tells how long this track will play
  return 1.0;
};
overrider bool Voice_Track_Class::keepgoing(){// tells when we have passed the end of duration.
  return false;
  // return keeprunning;
};
//-------------------------------------------------------------------------------------
overrider Time_Slicer_Class* Voice_Track_Class::spawn_slicer(){// creates a "read head" to keep a time marker in this playable
  Voice_Track_Slicer_Class *vs = new Voice_Track_Slicer_Class(this);
  //vs->init_slicer(this->get(0)->time, 0, this->size()-1);
  // probably should transfer time through the slicer class
  // 1 selected object sets the slicer's time to wherever the object's time starts?
  // 2 the player gets its time from the slicer at every moment?
  vs->init_slicer(vs->time_base_abs);
  // vs->init_slicer(this->time);
  return vs;
};
//-------------------------------------------------------------------------------------
size_t Voice_Track_Class::Tree_Search(double keytime){// Tree search for the first place where this keytime would go in order.
  size_t minloc=0, maxloc=this->size(), medloc;
  while (maxloc>minloc){
    /* medloc is the average between min and max. */
    medloc=(minloc+maxloc)>>1; /* >>1 is same as div 2, only faster. */
    Event_Marker_Class *cpoint = this->get(medloc);
    if (keytime<=cpoint->time){maxloc=medloc;}/* key is less than or equal to this list entry. */
    else{ minloc=medloc+1; }
  }
  return(minloc);
};/* Tree_Search */

/* cloning?
 http://www.new-brunswick.net/workshop/c++/faq/virtual-functions.html
 Voice_Track_Class* Voice_Track_Class::clone()  const { return new Voice_Track_Class(*this); }*/

/* ************************************************************************************************************ */
/* Drawable_Class functions */
//-------------------------------------------------------------------------------------
overrider void Voice_Track_Class::Draw_Me(wxDC& dc, Box_Mapper_Class *Box_Map){/* do we need to know scale or box_mapper to do this? wxDC& dc */
  wxPen pen(*wxRED, 1); // red pen of width 1
  dc.SetPen(pen);

  //handlex=this->get(0)->time;handley=this->get(0)->octave;
  handlex=this->time; handley=this->octave;

  Draw_Handle(dc, Box_Map);

  wxBrush pink_brush; pink_brush.SetColour(255, 0, 255);
  wxBrush red_brush; red_brush.SetColour(255, 0, 0);
  wxBrush grey_brush; grey_brush.SetColour(128, 128, 128);
  dc.SetBrush(pink_brush);

  dc.SetPen(wxNullPen);
  //-------------------------------------------------------------------------------------
  double loudness_scale = 20.0;
  int start_dex=0; int end_dex=this->size()-1;
  wxPoint points[4];
  Event_Marker_Class *event_prev = this->get(start_dex);
  double x_prev, y_prev;
  double vol_prev = this->loudness * (event_prev->loudness*loudness_scale);

  Box_Map->Map_To_Screen(this->time+event_prev->time, this->octave+event_prev->octave, &x_prev, &y_prev);// makes this relative to my base time.
  for (int cnt=start_dex+1;cnt<=end_dex;cnt++){// inclusive
    Event_Marker_Class *event_next = this->get(cnt);
    double x_next, y_next;
    double vol_next = this->loudness * (event_next->loudness*loudness_scale);
    Box_Map->Map_To_Screen(this->time+event_next->time, this->octave+event_next->octave, &x_next, &y_next);// makes this relative to my base time.

    dc.SetPen(*wxBLACK_PEN); dc.SetBrush(grey_brush);

    points[0].x=(int)x_prev; points[0].y=(int)(y_prev-vol_prev);
    points[1].x=(int)x_next; points[1].y=(int)(y_next-vol_next);
    points[2].x=(int)x_next; points[2].y=(int)(y_next+vol_next);
    points[3].x=(int)x_prev; points[3].y=(int)(y_prev+vol_prev);
    dc.DrawPolygon(4, points);

    /* plot prev point so we paint *over* the already-drawn grey area */
    dc.SetBrush(red_brush); dc.DrawCircle((wxCoord)x_prev, (wxCoord)y_prev, (wxCoord)(event_prev->handle_radius));
    event_prev = event_next; x_prev = x_next; y_prev = y_next; vol_prev = vol_next;/* rollover */
  }
  dc.SetBrush(red_brush); dc.DrawCircle((wxCoord)x_prev, (wxCoord)y_prev, (wxCoord)(event_prev->handle_radius));/* last point */
};/* Voice_Track_Class::Draw_Me */
//-------------------------------------------------------------------------------------
overrider class Playable_Class* Voice_Track_Class::Deliver_Playable(double xloc, double yloc, Box_Mapper_Class *Box_Map){
  printf("Voice_Track_Class::Deliver_Playable");// xloc and yloc are screen coordinates
  Event_Marker_Class *pnt = NULL;
  //Event_Marker_Class *pnt_found = NULL;
  Playable_Class *play_found = NULL;
  double pntxloc, pntyloc;
  int num_points = this->size();
  for (int cnt=0;cnt<num_points;cnt++){
    pnt = this->get(cnt);
#if 0
    play_found = pnt->Deliver_Playable(xloc, yloc, Box_Map);// delivers self if hit
    if (play_found!=NULL){ break; }
#else
    // problem is that this only maps from my local coords to compare with absolute mouse coords.
    // what if I have parents of my own?  xloc and yloc are screen coords, might be hard to subtract local coords on each nested call.
    Box_Map->Map_To_Screen(this->time+pnt->time, this->octave+pnt->octave, &pntxloc, &pntyloc);
    // Box_Map->Map_To_Screen(pnt->time, pnt->octave, &pntxloc, &pntyloc);
    if (Math.Hypot(pntxloc-xloc,pntyloc-yloc) <= pnt->handle_radius){
      play_found = pnt; break;
    }
#endif
  }
  if (play_found == NULL){// if it isn't my children, maybe it's me
    Box_Map->Map_To_Screen(this->handlex, this->handley, &pntxloc, &pntyloc);
    //pntxloc-=this->handle_radius;
    if (Math.Hypot(pntxloc-xloc,pntyloc-yloc) <= this->handle_radius){play_found = this;}
  }
  return play_found;
};
/* **********************************************************************************************************************
  End Voice_Track_Class methods
********************************************************************************************************************** */
//-------------------------------------------------------------------------------------
Voice_Track_Class::Voice_Track_Slicer_Class::Voice_Track_Slicer_Class(Voice_Track_Class *new_timeline):Time_Slicer_Class(new_timeline){
  Voice_Track = new_timeline;
  sgen = Voice_Track->sgen;
  event0_dex=0; event1_dex=0;
};
//-------------------------------------------------------------------------------------
overrider void Voice_Track_Class::Voice_Track_Slicer_Class::init_slicer(double start_time){
  keeprunning = true;
  Establish_Location();
  int first_cpoint = Voice_Track->Tree_Search(start_time);
  if (first_cpoint<Voice_Track->size()){
    start_dex = first_cpoint; end_dex = Voice_Track->size()-1;
    event0_dex=start_dex; event1_dex=start_dex+1;
    event0 = Voice_Track->get(event0_dex);
    event1 = Voice_Track->get(event1_dex);
    sgen->clear();
    local_time = 0;// local time inside this block
  }
};
//-------------------------------------------------------------------------------------
overrider double Voice_Track_Class::Voice_Track_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  double returnval = 0.0;
  // this should be based on
  // lots of effort for a simple loop - simplify please
  double playtime = nowtime - this->my_playable->time;// makes this relative to my playable's base time.
  // double playtime = nowtime - this->time_abs;// abs time of the one who spawned me.
  if (playtime >= event1->time){// iterators
    if (event1_dex<end_dex){// rollover
      //printf("event1_dex:%i:\n",event1_dex);
      event0 = event1; event0_dex=event1_dex;
      event1_dex++;
      event1 = Voice_Track->get(event1_dex);
    }else{
      // then one full iteration has passed, jump out.
      returnval = 0.0;
      return returnval;
    }
  }
  {// meat of computation
    // yeesh, playtime is offset, do the events need slicers of their own to maintain their offsets?
    event0->intep_to_next_octave(event1, playtime, sgen);//event1==baadf00d;//need some range protection somewhere.
    // double amp = event0->intep_to_next_octave(event1, time, sgen);
    returnval = sgen->getamp(playtime);
  }
  return returnval;
};
//-------------------------------------------------------------------------------------
/* Playable_Class members */
/* **********************************************************************************************************************
  Begin Event_Marker_Class methods and callback
********************************************************************************************************************** */
Event_Marker_Class::Event_Marker_Class(){//:Playable_Class(){
  my_voice = NULL;// new_voice;
  this->time=0.0; this->set_frequency(440.0); this->loudness=1.0;;// defaults
  audible = true;// keep playing from this event to the next
  handle_radius = 5;
};
//-------------------------------------------------------------------------------------
void Event_Marker_Class::intep_to_next_frequency(Event_Marker_Class *nextmarker, double absolutetime, Signal_Generator_Class *sgen){
  double fullspan = nextmarker->time - this->time;
  //if (fullspan==0.0){fullspan=fudge;}// prevent divide-by-zero.
  double delta = absolutetime - this->time;
  double weight_next = delta/fullspan;
  double weight_prev = 1.0 - weight_next;

  double interloudness = (this->loudness * weight_prev) + (nextmarker->loudness * weight_next);
  double interfreq = (this->frequency * weight_prev) + (nextmarker->frequency * weight_next);
  // System.out.println(weight_prev);
  sgen->freq = interfreq; sgen->loudness = interloudness;
  // System.out.println("f:"+interfreq+" t:"+absolutetime);
  // System.out.println(" "+interfreq+", "+absolutetime);
  if (!audible){
    sgen->loudness = 0.0;
  }
};
//-------------------------------------------------------------------------------------
void Event_Marker_Class::intep_to_next_octave(Event_Marker_Class *nextmarker, double absolutetime, Signal_Generator_Class *sgen){
  double fullspan = nextmarker->time - this->time;
  double delta = absolutetime - this->time;
  double weight_next = delta/fullspan;
  double weight_prev = 1.0 - weight_next;
  double interloudness = (this->loudness * weight_prev) + (nextmarker->loudness * weight_next);
  double interoctave = (this->octave * weight_prev) + (nextmarker->octave * weight_next);
  sgen->set_octave(interoctave); sgen->loudness = interloudness;
  if (!audible){
    sgen->loudness = 0.0;
  }
};
overrider double Event_Marker_Class::getamp(double time){// only used for jugar
  Signal_Generator_Class *sgen = this->my_voice->sgen;
  sgen->set_octave(this->octave);
  double amp = sgen->getamp(time) * this->loudness;
  return amp;
};
/* these are all of the playable_class functions, which are only if ever used during jugar */
overrider double Event_Marker_Class::maxamp(){return 0.0;};
overrider double Event_Marker_Class::get_duration(){return 0.0;};// tells how long this track will play
overrider bool Event_Marker_Class::keepgoing(){return 0.0;};// tells when we have passed the end of duration.
overrider class Time_Slicer_Class* Event_Marker_Class::spawn_slicer(){
  return new Event_Marker_Class::Event_Marker_Slicer_Class(this);
};// creates a "read head" to keep a time marker in this playable
overrider void Event_Marker_Class::Draw_Me(wxDC& dc, Box_Mapper_Class *Box_Map){
  this->my_voice->Draw_Me(dc, Box_Map);
};
overrider class Playable_Class* Event_Marker_Class::Deliver_Playable(double xloc, double yloc, Box_Mapper_Class *Box_Map){
  printf("Event_Marker_Class::Deliver_Playable");// xloc and yloc are screen coordinates
  double pntxloc, pntyloc;
  Box_Map->Map_To_Screen(this->time, this->octave, &pntxloc, &pntyloc);
  Event_Marker_Class *pnt_found = NULL;
  if (Math.Hypot(pntxloc-xloc,pntyloc-yloc) <= this->handle_radius){
    pnt_found = this;// return this;
  }else{
    pnt_found = NULL;// return NULL;
  }
  return pnt_found;
};
/* **********************************************************************************************************************
  End Event_Marker_Class methods and callback
********************************************************************************************************************** */
/* **********************************************************************************************************************
  Begin Event_Marker_Slicer_Class methods and callback
********************************************************************************************************************** */
Event_Marker_Class::Event_Marker_Slicer_Class::Event_Marker_Slicer_Class(class Event_Marker_Class *New_Event_Marker):Time_Slicer_Class(New_Event_Marker){
  My_Event_Marker = New_Event_Marker;
};
overrider void Event_Marker_Class::Event_Marker_Slicer_Class::init_slicer(double start_time){// this one is different for every descendent
};
overrider double Event_Marker_Class::Event_Marker_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  // only used for jugar
  //cpoint->spawn_slicer
  //nowtime -= xloc; nowoctave += yloc;/* yloc is base octave for this frame */ nowvol *= my_playable->zloc;
  //nowtime -= my_playable->xloc; nowoctave += my_playable->yloc;/* yloc is base octave for this frame */ nowvol *= my_playable->zloc;
  nowoctave += this->My_Event_Marker->octave;// my_playable->octave;
  Signal_Generator_Class *sgen = this->My_Event_Marker->my_voice->sgen;
  //nowoctave = Math.Log(440)/Math.Log(2.0);
  sgen->set_octave(nowoctave);
  sgen->loudness = this->My_Event_Marker->loudness;
  double amp = sgen->getamp(nowtime) * nowvol;
  return amp;
  //x;// issue is 1. how do we offset this for its base origins?  2. intep_to_next_octave with self?
};
overrider bool Event_Marker_Class::Event_Marker_Slicer_Class::keepgoing(){// tells when we have passed the end of duration.
  return keeprunning;
};
/* **********************************************************************************************************************
  End Event_Marker_Slicer_Class methods and callback
********************************************************************************************************************** */
