/***************************************************************
 * Name:      drawing_canvas.cpp
 * Purpose:   For drawing music
 * Author:    John Conner Atkeson
 * Created:   2007-11-06
 * Copyright: John Conner Atkeson (http://jca3.freeshell.org)
 * License:   LGPL
 **************************************************************/

#include "drawing_canvas.h"
#define TIMER_ID 122
DrawingCanvasClass::DrawingCanvasClass(wxWindow *parent, int id) : wxPanel(parent, id, wxDefaultPosition, wxSize(-1, 30), wxSUNKEN_BORDER)
{
  // : m_timer(this, TIMER_ID) m_timer.Start(1000);    // 1 second interval
  m_parent = parent;
  this->SetSize(500, 300);

  // const long ID_PANEL_THIS = wxNewId();  Connect(ID_PANEL_THIS,wxEVT_PAINT,(wxObjectEventFunction)&DrawingCanvas::OnPaint,0,this);
  Connect(wxEVT_PAINT, wxPaintEventHandler(DrawingCanvasClass::OnPaint));

  Connect(wxEVT_SIZE, wxSizeEventHandler(DrawingCanvasClass::OnSize));
  Connect(wxEVT_MOTION, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelMouseMove));

  Connect(wxEVT_LEFT_DOWN, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelLeftDown));
  Connect(wxEVT_LEFT_UP, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelLeftUp));

  Connect(wxEVT_RIGHT_DOWN, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelRightDown));
  Connect(wxEVT_RIGHT_UP, wxMouseEventHandler(DrawingCanvasClass::OnDrawingPanelRightUp));

  // Connect(wxEVT_COMMAND_BUTTON_CLICKED, wxSizeEventHandler(DrawingCanvasClass::OnQuit));
  // Connect(ID_BUTTON2,wxEVT_COMMAND_BUTTON_CLICKED,(wxObjectEventFunction)&DrawingCanvasClass::OnQuit);
  // void OnShow(wxEvent& event); EVT_SHOW (MyFrame::OnShow);
  counter = 0;
  left_mouse_down = true;
  altstart = 0.0;
  loudness_start = 0.0;
#if have_player
  // Reffy_Ptr_Class sgen_ptr;
  sgen = new Freewave_Generator_Class(440.0);

  sgen->freq = 440;
  player = new Player_Class();
#endif

  Root_Playable = new Playable_Tree_Class();
  //this->SetSize(300, 301);
  Voice_Track = new Voice_Track_Class(sgen);//new Freewave_Generator_Class(220.0)
  Voice_Track->fill_test();

  Root_Playable->kids->add(Voice_Track);

  //player->slicer = sgen->spawn_slicer();
  //player->slicer = Voice_Track->spawn_slicer();
  player->slicer = NULL;
  Selector.Set_Selection(NULL);

  Box_Map = new Box_Mapper_Class();
  wxSize sz = GetClientSize();
  Box_Map->Set_Screen(0, sz.GetHeight(), sz.GetWidth(), 0);
  //Box_Map->Set_Real(0.0,5.0, ((double)sz.GetWidth())/70.0, 12.0);
  Box_Map->Set_Real(0.0,0.0, ((double)sz.GetWidth())/70.0, 22.0);

  Trackline_Timer = new Drawing_Canvas_Timer(this);
#if 0
  m_timer(this, TIMER_ID);
  m_timer.Start(1000, wxTIMER_CONTINUOUS);
#endif
  wxSize clsz = this->GetClientSize();
  int wdt = clsz.GetWidth(); int hgt = clsz.GetHeight();
  Background_Pic = new wxBitmap(wdt, hgt);
  Gradient_Backdrop = new wxBitmap(wdt, hgt);
  Fill_Gradient_Backdrop();

  /* Draw to a memory device. */
  Repaint_Background();

  Trackline_Prev_Time=0.0; Trackline_Now_Time=0.0;
};
#ifdef etable
//IMPLEMENT_CLASS(DrawingCanvasClass, wxPanel)
BEGIN_EVENT_TABLE(DrawingCanvasClass, wxPanel)
  // EVT_IDLE(DrawingCanvasClass::OnIdle)
  EVT_TIMER(TIMER_ID, DrawingCanvasClass::OnTimer)
END_EVENT_TABLE()
#endif
/* ************************************************************************************************************ */
DrawingCanvasClass::~DrawingCanvasClass(){
  Trackline_Timer->Stop();
  delete Background_Pic;
  delete Gradient_Backdrop;
  delete Voice_Track;
  delete Root_Playable;
  delete Box_Map;
#if have_player
  delete player;
  delete sgen;
#endif
  delete Trackline_Timer;
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::Repaint_Background(){
  /* Draw to a memory device. */
  wxMemoryDC mem_dc; mem_dc.SelectObject((*Background_Pic));
  Painter(mem_dc);
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::Fill_Gradient_Backdrop(){
  // http://www.wxwidgets.org/manuals/2.6/wx_wxbitmap.html
  wxMemoryDC mem_dc; mem_dc.SelectObject((*Gradient_Backdrop));
  wxRect rect(0,0, Gradient_Backdrop->GetWidth(), Gradient_Backdrop->GetHeight());
  // mem_dc.GradientFillLinear(rect, '#ffec00', '#000000', wxNORTH);
  //wxColour(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha=wxALPHA_OPAQUE)

  wxColor color0(0xff, 0xff, 0xff);//, wxALPHA_OPAQUE
  wxColor color1(0x00, 0xf0, 0xff);//, wxALPHA_OPAQUE

  int num_colors = 3;
  wxColor **spectrum;
  spectrum = allocsafe(num_colors,wxColor*);
  for (int cnt=0;cnt<num_colors;cnt++){
    spectrum[cnt] = new wxColor();
  }

  int alpha = wxALPHA_TRANSPARENT;//0x88;// wxALPHA_OPAQUE;
  int pale = 0x88;
  spectrum[2]->Set(0xff, pale, pale, alpha);// red
  spectrum[1]->Set(0xff, 0xff, pale, alpha);// yellow
  spectrum[0]->Set(pale, pale, 0xff, alpha);// blue

  if (true){
    double xreal0, yreal0, xreal1, yreal1;
    Box_Map->Map_To_Real(0, 0, &xreal1, &yreal1);// highest octave
    Box_Map->Map_To_Real(Gradient_Backdrop->GetWidth(), Gradient_Backdrop->GetHeight(), &xreal0, &yreal0);// lowest octave
    yreal0 = Math.Floor(yreal0); yreal1 = Math.Ceil(yreal1);
    wxRect spectrect;//(0,0, Gradient_Backdrop->GetWidth(), Gradient_Backdrop->GetHeight());
    /*
    first we need to map 1 octave to the height.  then create height.
    next we need to map the limits of the box to octaves.
    we either iterate with octaves, or we iterate with real box measurements.
    either way the iterations must go beyond the edges of the scene right?
    map box bottom to octaves, round down. Math.Floor()
    map box top to octaves, round up.  Math.Cieling()
    now we can iterate from lowest to highest octave, placing a rectangle at each to_screen mapped point.
    or we can map lowest and highest octaves back to box, and iterate that way (no, how much is 1 step?)
    lowest becomes first
    */

    double xloc, yloc;
    Box_Map->Map_To_Screen(0.0, 0.0, &xloc, &yloc);
    double baseoct = yloc;
    Box_Map->Map_To_Screen(0.0, 1.0, &xloc, &yloc);
    wxSize rsz(Gradient_Backdrop->GetWidth(), (int)(baseoct-yloc));
    rect.SetSize(rsz);

    wxSize spectsz(Gradient_Backdrop->GetWidth(), (int)(1.0 + (baseoct-yloc)/((double)num_colors)));
    spectrect.SetSize(spectsz);
    for (double octcnt=yreal0;octcnt<=yreal1;octcnt++){
      Box_Map->Map_To_Screen(0, octcnt, &xloc, &yloc);
      rect.SetY(yloc);
      //mem_dc.GradientFillLinear(rect, color0, color1, wxNORTH);
      for (int scnt=0;scnt<num_colors;scnt++){
        spectrect.SetY(yloc + (scnt*spectrect.GetHeight()));
        int next_color = (scnt+1)%num_colors;
        mem_dc.GradientFillLinear(spectrect, *(spectrum[next_color]), *(spectrum[scnt]), wxNORTH);
      }
    }
  }else{
    mem_dc.GradientFillLinear(rect, color0, color1, wxNORTH);
  }
  //mem_dc.GradientFillLinear(rect, *wxRED, *wxBLUE, wxNORTH);
  //
  for (int cnt=0;cnt<num_colors;cnt++){ delete spectrum[cnt]; }
  freesafe(spectrum);
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::Painter(wxDC& dc){
  // http://www.informit.com/articles/printerfriendly.aspx?p=405047&rl=1
  {// blank background
    dc.SetBrush(*wxWHITE_BRUSH);
    dc.Clear();

    wxMemoryDC mem_dc; mem_dc.SelectObject((*Gradient_Backdrop));
    dc.Blit(/* xdest */0, /* ydest */0, /* wdt */Gradient_Backdrop->GetWidth(), /* hgt */Gradient_Backdrop->GetHeight(), &mem_dc, /* xsrc */0, /* ysrc */0);
  }
  wxPen pen(*wxRED, 1); // red pen of width 1
  dc.SetPen(pen);

  wxBrush pink_brush; pink_brush.SetColour(255, 0, 255);
  wxBrush red_brush; red_brush.SetColour(255, 0, 0);
  wxBrush grey_brush; grey_brush.SetColour(128, 128, 128);
  dc.SetBrush(pink_brush);

  dc.SetPen(wxNullPen);
  if (false){
    dc.SetPen(*wxBLACK_PEN);
    // Get window dimensions
    wxSize sz = GetClientSize();

    // Our rectangle dimensions
    wxCoord w = 100, h = 50;
    // Center the rectangle on the window, but never draw at a negative position.
    int x = wxMax(0, (sz.x-w)/2);
    int y = wxMax(0, (sz.y-h)/2);
    wxRect rectToDraw(x, y, w, h);
    // dc.DrawLine(x, y, w, h);
    // For efficiency, do not draw if not exposed
    if (IsExposed(rectToDraw)){
      dc.DrawRectangle(rectToDraw);
    }
  }
  //-------------------------------------------------------------------------------------
  Voice_Track->Draw_Me(dc, Box_Map);
};/* DrawingCanvasClass::Painter */
/* ************************************************************************************************************ */
void DrawingCanvasClass::Draw_Trackline(){
  if (this->player!=NULL){
    double xloc, yloc = 0.0;
    double linehgt;
    wxClientDC dc(this);
    /* cover the previous position from the image buffer */
    Box_Map->Map_To_Screen(Trackline_Prev_Time , yloc, &xloc, &yloc);
    wxSize clsz = this->GetClientSize(); int wdt = clsz.GetWidth(); int hgt = clsz.GetHeight();
    linehgt = hgt;//Background_Pic->GetWidth(), Background_Pic->GetHeight(),
    if (0<=xloc && xloc<wdt){
      wxMemoryDC mem_dc;
      mem_dc.SelectObject((*Background_Pic));
      dc.Blit(/* xdest */xloc, /* ydest */0, /* wdt */2, /* hgt */linehgt, &mem_dc, /* xsrc */xloc, /* ysrc */0);
    }
    Trackline_Prev_Time = Trackline_Now_Time;
    Trackline_Now_Time = this->player->time;
    Box_Map->Map_To_Screen(Trackline_Now_Time , yloc, &xloc, &yloc);
    dc.DrawLine((wxCoord)xloc, (wxCoord)(0), (wxCoord)xloc, (wxCoord)(linehgt));
  }
};/* Draw_Trackline */
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnIdle(wxIdleEvent &event){
  //printf("counter:%li\n",counter);
  if (counter==0){
    //this->Draw_Trackline();
  }
  counter++;
  if (counter>100) {counter=0;}
  event.RequestMore();
  //event.Skip();
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnTimer(wxTimerEvent& event){
  printf("hello timer");
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelRightDown(wxMouseEvent& event)
{
  printf("OnDrawingPanelRightDown");
  double xloc = event.GetX();
  double yloc = event.GetY();
  // double xray, yray; Box_Map->Map_To_Real(xloc,yloc, &xray,&yray);

  player->quit();
  Selector.Set_Selection(this->Voice_Track);
  if (player->slicer!=NULL){
    delete player->slicer; player->slicer = NULL;
  }
  if (Selector.selection!=NULL){
    player->slicer = Selector.selection->spawn_slicer();
    player->sdlsoundinit();
    player->time = player->slicer->time_abs;
  }
  player->time = 0.0;
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelRightUp(wxMouseEvent& event)
{
  printf("OnDrawingPanelRightUp");
  double xloc = event.GetX();
#if have_player
  player->quit();
  if (player->slicer!=NULL){
    delete player->slicer; player->slicer = NULL;
  }
#endif
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnPaint(wxPaintEvent& event){
  wxPaintDC dc(this);
  wxMemoryDC mem_dc; mem_dc.SelectObject((*Background_Pic));
  dc.Blit(0, 0, Background_Pic->GetWidth(), Background_Pic->GetHeight(), &mem_dc, 0, 0);
  // Painter(dc);
}/* DrawingCanvasClass::OnPaint */
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnSize(wxSizeEvent& event)
{
  wxSize clsz = this->GetClientSize();
  if (false){
    Box_Map->Set_Screen(0, clsz.GetHeight(), clsz.GetWidth(), 0);// inverted Y axis
    Box_Map->Set_Real(0,7, ((double)clsz.GetWidth())/70.0, 12.0);
  }
  delete Gradient_Backdrop;
  delete Background_Pic;
  int wdt = clsz.GetWidth(); int hgt = clsz.GetHeight();
  Gradient_Backdrop = new wxBitmap(wdt, hgt);
  Background_Pic = new wxBitmap(wdt, hgt);

  Refresh();
  printf("DrawingCanvasClass::OnSize");
};/* DrawingCanvasClass::OnSize */
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelLeftDown(wxMouseEvent& event)
{// from Waveform_Canvas_Class
  printf("OnDrawingPanelLeftDown");
  left_mouse_down = true;
  bool altdown = event.AltDown();
  double xloc = event.GetX();
  double yloc = event.GetY();
  //double xray, yray; Box_Map->Map_To_Real(xloc,yloc, &xray,&yray);
#if have_player
  player->quit();
  if (true){
    // Root_Playable needs to be a generic container of some kind.
    // Selector.Set_Selection(Root_Playable->Deliver_Playable(xloc, yloc, Box_Map));
    Selector.Set_Selection(Voice_Track->Deliver_Playable(xloc, yloc, Box_Map));
    if (player->slicer!=NULL){
      player->quit();
      delete player->slicer; player->slicer = NULL;
    }
    if (Selector.selection!=NULL){
      player->slicer = Selector.selection->spawn_slicer();
      player->sdlsoundinit();
      player->slicer->Establish_Location();
      player->time = player->slicer->time_abs;

      if (altdown){
        altstart = event.GetY();
        loudness_start = Selector.selection->loudness;
      }
    }
  }
#endif
};/* DrawingCanvasClass::OnDrawingPanelLeftDown */
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelLeftUp(wxMouseEvent& event)
{
  left_mouse_down = false;
#if have_player
  player->quit();
  if (player->slicer!=NULL){
    delete player->slicer; player->slicer = NULL;
  }
#endif
};
/* ************************************************************************************************************ */
void DrawingCanvasClass::OnDrawingPanelMouseMove(wxMouseEvent& event)
{
  //event.RightDown
  //bool buttonstate = event.GetButton();
  bool bdown_left = event.ButtonIsDown(wxMOUSE_BTN_LEFT);
  bool bdown_right = event.ButtonIsDown(wxMOUSE_BTN_RIGHT);
  bool altdown = event.AltDown();
  //bool dragging = event.Dragging();
  double xloc = event.GetX();
  double yloc = event.GetY();
  // bool buttonstate = event.Button(0);
  if (bdown_left){
    /*
    here we need to pick the exact event point of the voice and play it in a loop while we move it.
    alternatively we will be picking whole groups (any playable) and playing the whole group in a loop.
    */
    printf("left_mouse_down move!\n");
    if (Selector.selection!=NULL){
      double new_time,new_octave;
      double xorg, yorg, zorg;
      xorg = Selector.time_base_abs; yorg = Selector.octave_base_abs; zorg = Selector.loudness_base_abs;
      Box_Map->Map_To_Real(xloc,yloc, &new_time, &new_octave);
      new_time-=xorg; new_octave-=yorg; //zorg

      Selector.selection->time = new_time;
      if (altdown){
        double new_loudness = new_octave-(Selector.selection->octave);
        if (new_loudness<0.0){new_loudness=0.0;} else if (new_loudness>1.0){ new_loudness=1.0; }
        Selector.selection->loudness = new_loudness;
      }else{
        Selector.selection->set_octave(new_octave);
      }
      if (player->slicer!=NULL){// hack, this shouldn't ever really be null should it?
        player->slicer->Establish_Location();
      }
    }
    // handle graphics
    wxClientDC dc(this);
    wxSize clsz = this->GetClientSize();
    wxCoord w = 100, h = 50;
    int x = wxMax(0, (clsz.x-w)/2);
    int y = wxMax(0, (clsz.y-h)/2);
    {
      // alternative way
      // wxBufferedDC bufdc(&dc, clsz, wxBUFFER_CLIENT_AREA);
      // Painter(bufdc); bufdc.DrawLine((wxCoord)x, (wxCoord)y,  (wxCoord)xloc, (wxCoord)yloc);
      int wdt = clsz.GetWidth(); int hgt = clsz.GetHeight();
      int bwdt = wdt, bhgt = hgt;

      // Draw to a memory device.
      wxMemoryDC mem_dc; mem_dc.SelectObject((*Background_Pic));// Repaint_Background();
      Painter(mem_dc);
      int xsrc = 0, ysrc = 0, xdest = 0, ydest = 0;
      dc.Blit(xdest, ydest, bwdt, bhgt, &mem_dc, xsrc, ysrc);
    }
  }else{
    // printf("left_mouse_up move!\n");
  }
};/* DrawingCanvasClass::OnDrawingPanelMouseMove */
/* **********************************************************************************************************************
  Drawing_Canvas_Timer functions
********************************************************************************************************************** */
Drawing_Canvas_Timer::Drawing_Canvas_Timer(DrawingCanvasClass *new_canvas){
  this->my_canvas = new_canvas;
  this->Start(30, wxTIMER_CONTINUOUS);// 1000 is a full second
};
overrider void Drawing_Canvas_Timer::Notify(){
  my_canvas->Draw_Trackline();
};
#if 0
void DnDWindow::OnLeftDown(wxMouseEvent& event )
{
  if ( !m_strText.IsEmpty() )
  {
    // start drag operation
    wxTextDataObject textData(m_strText);
    wxDropSource source(textData, this, wxDROP_ICON(dnd_copy), wxDROP_ICON(dnd_move), wxDROP_ICON(dnd_none));
    int flags = 0;
    if ( m_moveByDefault )
      flags |= wxDrag_DefaultMove;
    else if ( m_moveAllow )
      flags |= wxDrag_AllowMove;
    wxDragResult result = source.DoDragDrop(flags);
    const wxChar *pc;
    switch ( result )
    {
    case wxDragError: pc = wxT("Error!"); break;
    case wxDragNone: pc = wxT("Nothing"); break;
    case wxDragCopy: pc = wxT("Copied"); break;
    case wxDragMove: pc = wxT("Moved"); break;
    case wxDragCancel: pc = wxT("Cancelled"); break;
    default: pc = wxT("Huh?"); break;
    }
    wxMessageBox(wxString(wxT("Drag result: ")) + pc);// old quotes:
  }
}
// A custom drop target that accepts URL objects
class URLDropTarget : public wxDropTarget
{
public:
  URLDropTarget() { SetDataObject(new wxURLDataObject); }
  void OnDropURL(wxCoord x, wxCoord y, const wxString& text)
  {
// Of course, a real program would do something more
// useful here...
    wxMessageBox(text, wxT("URLDropTarget: got URL"),
                 wxICON_INFORMATION | wxOK);
  }
// URLs cant be moved, only copied
  virtual wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def)
  {
    return wxDragLink;
  }
// translate this to calls to OnDropURL() just for convenience
  virtual wxDragResult OnData(wxCoord x, wxCoord y, wxDragResult def)
  {
    if ( !GetData() )
      return wxDragNone;
    OnDropURL(x, y, ((wxURLDataObject *)m_dataObject)->GetURL());
    return def;
  }
};
// Set the drop target
wxListBox* listBox = new wxListBox(parent, wxID_ANY);
listBox->SetDropTarget(new URLDropTarget);

#endif
