/***************************************************************
 * Name:      waveform_canvas.h
 * Purpose:   For drawing waveforms
 * Author:    John Conner Atkeson
 * Created:   2007-11-06
 * Copyright: John Conner Atkeson (http://jca3.freeshell.org)
 * License:   LGPL
 **************************************************************/

#ifndef waveform_canvas_h
#define waveform_canvas_h 1

#include "wx_pch.h"
#include <wx/msgdlg.h>

//(*InternalHeaders(fretlessDialog)
#include <wx/bitmap.h>
#include <wx/font.h>
#include <wx/fontenum.h>
#include <wx/fontmap.h>
#include <wx/image.h>
#include <wx/intl.h>
#include <wx/settings.h>
#include <wx/string.h>
#include <wx/dcbuffer.h>
#include <wx/dcmemory.h>
#include "audio.h"
#include "drawing_canvas.h"

#define have_player 1
/* ************************************************************************************************************ */
class Waveform_Canvas_Class : public wxPanel{
public:
  bool left_mouse_down;
#if have_player
  Player_Class *player;
#endif
  class Waveform_Editor_Class *My_Waveform_Editor;
  // wxSlider* My_Octave_Slider;
  wxPanel* DrawingPanel;
  wxWindow *m_parent;

  Voice_Track_Class *Voice_Track;
  Event_Marker_Class *cpoint;
  Reffy_Ptr_Class sgen_ptr;
  Splinewave_Generator_Class *sgen;
  Splinewave_Generator_Class::Point_Class *selected_point;
  double altstart;//, loudness_start;
  class Box_Mapper_Class *Box_Map, *Spline_Box_Map;
  Waveform_Canvas_Class::Waveform_Canvas_Class(Waveform_Editor_Class *parent, int id);
  /* ************************************************************************************************************ */
  Waveform_Canvas_Class::~Waveform_Canvas_Class();
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::OnShow(wxEvent& WXUNUSED(event)){
    // wxPaintDC dc(this);
    // Painter(dc);
    // wxMessageBox("Show event handled", _T("Shown"), wxOK | wxICON_INFORMATION, this);
  }
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::Painter(wxDC& dc);
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::OnPaint(wxPaintEvent& event){
    wxPaintDC dc(this);
    Painter(dc);
  }/* Waveform_Canvas_Class::OnPaint */
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::OnSize(wxSizeEvent& event)
  {
    Refresh();
  }/* Waveform_Canvas_Class::OnSize */
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::OnDrawingPanelLeftDown(wxMouseEvent& event);
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::OnDrawingPanelLeftUp(wxMouseEvent& event){
    left_mouse_down = false;
#if have_player
    player->quit();
#endif
  }
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::OnDrawingPanelMouseMove(wxMouseEvent& event);
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::OnQuit(wxCommandEvent& event)
  {
    Close();
  };
  /* ************************************************************************************************************ */
  void Waveform_Canvas_Class::Link_Signal_Generator(Splinewave_Generator_Class *new_sgen){
    //if (sgen!=NULL){delete sgen; }
    sgen = new_sgen;//Reffy_Ptr_Class sgen_ptr;
  };
  /* ************************************************************************************************************ */
  double realleft, realright, realwdt;
  double realtop, realbottom, realhgt;
  double scrleft, scrright, scrwdt;
  double scrtop, scrbottom, scrhgt;
  void Map_To_Screen(double xreal, double yreal, double *xscr, double *yscr){
    double xnew = (xreal-realleft)/realwdt;
    double ynew = (yreal-realtop)/realhgt;
    (*xscr) = scrleft + (xnew * scrwdt);
    (*yscr) = scrtop + (ynew * scrhgt);
  }
};/* Waveform_Canvas_Class */
/* ************************************************************************************************************ */
class Waveform_Editor_Class : public wxPanel{
public:
  wxBoxSizer *BoxSizer;
  wxSlider *Octave_Slider;
  Waveform_Canvas_Class *Waveform_Canvas;
  wxButton* Save_Button;

  Waveform_Editor_Class::Waveform_Editor_Class(wxWindow *parent, int id);
  void Waveform_Editor_Class::OnOctave_SliderCmdScrollChanged(wxScrollEvent& event);
  void Waveform_Editor_Class::OnOctave_SliderCmdScroll(wxScrollEvent& event);
  void Waveform_Editor_Class::OnOctave_SliderCmdScrollThumbTrack(wxScrollEvent& event);
  void Waveform_Editor_Class::OnOctave_SliderCmdScrollLeftDown(wxScrollEvent& event);
  void Waveform_Editor_Class::OnOctave_SliderCmdScrollThumbRelease(wxScrollEvent& event);

  void Waveform_Editor_Class::OnButton1Click(wxCommandEvent& event);
};
/* ************************************************************************************************************ */
#endif // waveform_canvas_h

#if 0 // junkyard
/*



next step is we want a button to save the wave form, as a list of xy points. {x,y}, {x,y}, {x,y};
and we also want a pitch shifter.;


quadratic bezier curve:
b =
+ sq(1-t)*P0
+ 2*t*(1-t)*P1
+ sq(t)*P2

quadratic formula where axx + bx + c = 0:
x =
-b +-
  sqrt(bb - 4ac)/
  2a


from http://www.quickmath.com/:
input: Z = A(1-t)^2 + 2Bt(1-t) + Ct^2

output:
t = (a - b - sqrt(b*b - 2zb - ac + az + cz)) / (a - 2b + c)
and:
t = (a - b + sqrt(b*b - 2zb - ac + az + cz)) / (a - 2b + c)

t = (p0 - p1 - sqrt(p1*p1 - 2*X*p1 - p0*p2 + p0*X + p2*X)) / (p0 - 2*p1 + p2)
and:
t = (p0 - p1 + sqrt(p1*p1 - 2*X*p1 - p0*p2 + p0*X + p2*X)) / (p0 - 2*p1 + p2)

then:
Y = sq(1-t)*YP0 + 2*t*(1-t)*YP1 + sq(t)*YP2

{
    double p0 = 2.2, p1 = 12.3, p2 = 15.6;
    double X,Y, t;
    for (X = p0; X <= p2; X++)
    {
        t = (p0 - p1 - Math.Sqrt(p1 * p1 - 2 * X * p1 - p0 * p2 + p0 * X + p2 * X)) / (p0 - 2 * p1 + p2);
        Y = Math.Pow(1 - t, 2) * p0 + 2 * t * (1 - t) * p1 + Math.Pow(t, 2) * p2;
        System.Console.WriteLine("X:{0}:  Y:{1}:   t:{2}:", X, Y, t);
    }
}
X:2.2:  Y:2.20000000000002:   t:2.97058823529412:
X:3.2:  Y:3.19999999999999:   t:2.92022958392683:
X:4.2:  Y:4.20000000000003:   t:2.86803811740194:
X:5.2:  Y:5.2:   t:2.81379784189051:
X:6.2:  Y:6.2:   t:2.75724668036286:
X:7.2:  Y:7.20000000000002:   t:2.69806140705995:
X:8.2:  Y:8.2:   t:2.63583558788639:
X:9.2:  Y:9.19999999999997:   t:2.5700460944653:
X:10.2:  Y:10.2:   t:2.5:
X:11.2:  Y:11.2:   t:2.42474566198144:
X:12.2:  Y:12.2:   t:2.34291313598961:
X:13.2:  Y:13.2:   t:2.2524000797497:
X:14.2:  Y:14.2:   t:2.14966805871038:
X:15.2:  Y:15.2:   t:2.02781961016755:

{
    double p0 = 2.2, p1 = 12.3, p2 = 15.6;
    double X,Y, t;
    for (X = p0; X <= p2; X++)
    {
        t = (p0 - p1 + Math.Sqrt(p1 * p1 - 2 * X * p1 - p0 * p2 + p0 * X + p2 * X)) / (p0 - 2 * p1 + p2);
        Y = Math.Pow(1 - t, 2) * p0 + 2 * t * (1 - t) * p1 + Math.Pow(t, 2) * p2;
        System.Console.WriteLine("X:{0}:  Y:{1}:   t:{2}:", X, Y, t);
    }
}
X:2.2:  Y:2.2:   t:0:
X:3.2:  Y:3.2:   t:0.0503586513672882:
X:4.2:  Y:4.2:   t:0.102550117892176:
X:5.2:  Y:5.2:   t:0.156790393403608:
X:6.2:  Y:6.2:   t:0.213341554931252:
X:7.2:  Y:7.2:   t:0.272526828234166:
X:8.2:  Y:8.2:   t:0.334752647407727:
X:9.2:  Y:9.2:   t:0.400542140828822:
X:10.2:  Y:10.2:   t:0.470588235294117:
X:11.2:  Y:11.2:   t:0.54584257331268:
X:12.2:  Y:12.2:   t:0.62767509930451:
X:13.2:  Y:13.2:   t:0.718188155544413:
X:14.2:  Y:14.2:   t:0.820920176583734:
X:15.2:  Y:15.2:   t:0.942768625126567:

public void test_math()
{
    double p0 = 2.2, p1 = 12.3, p2 = 15.6;
    double X, Y, t, comp_t;
    for (X = p0; X <= p2; X+=0.1415926)
    {
        t = (p0 - p1 + Math.Sqrt((p1 * p1) - (2 * X * p1) - (p0 * p2) + (p0 * X) + (p2 * X))) / (p0 - (2 * p1) + p2);
        comp_t = 1 - t;
        Y = (comp_t * comp_t * p0) + (2 * t * comp_t * p1) + (t * t * p2);
        System.Console.WriteLine("X:{0}:  Y:{1}:   t:{2}:", X, Y, t);
    }
    bool noop = true;
}
*/
#endif

